"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.OssDatastore = exports.createOssRepo = exports.OssLock = void 0;
const buffer_1 = require("buffer");
const it_filter_1 = __importDefault(require("it-filter"));
const interface_datastore_1 = require("interface-datastore");
const base_1 = require("datastore-core/base");
const Errors = __importStar(require("datastore-core/errors"));
const uint8arrays_1 = require("uint8arrays");
const it_to_buffer_1 = __importDefault(require("it-to-buffer"));
const ali_oss_1 = __importDefault(require("ali-oss"));
var ossLock_1 = require("./lib/ossLock");
Object.defineProperty(exports, "OssLock", { enumerable: true, get: function () { return ossLock_1.OssLock; } });
var create_oss_repo_1 = require("./lib/create-oss-repo");
Object.defineProperty(exports, "createOssRepo", { enumerable: true, get: function () { return create_oss_repo_1.createOssRepo; } });
/**
 * A datastore backed by the file system, save to aliyun oss.
 *
 * Keys need to be sanitized before use, as they are written
 * to the file system as is.
 */
class OssDatastore extends base_1.BaseDatastore {
    constructor(path, opts) {
        super();
        this.path = path
            // aliyun not suport /path/, use path/
            .replace(/^\//, "");
        this.opts = opts;
        const { createIfMissing = false, oss, ossOption } = opts;
        if (oss) {
            this.ossClient = oss;
        }
        else {
            if (!ossOption) {
                throw new Error("oss and ossOptions must have one. See the README for examples.");
            }
            if (typeof ossOption?.bucket !== "string") {
                throw new Error("ossOption.bucket must be supplied. See the README for examples.");
            }
            this.ossClient = new ali_oss_1.default({
                ...ossOption,
            });
        }
        if (typeof createIfMissing !== "boolean") {
            throw new Error(`createIfMissing must be a boolean but was (${typeof createIfMissing}) ${createIfMissing}`);
        }
        this.createIfMissing = createIfMissing;
    }
    path;
    opts;
    createIfMissing;
    ossClient;
    /**
     * Returns the full key which includes the path to the ipfs store
     */
    _getFullKey(key) {
        // Avoid absolute paths with oss
        this.opts.hasLog && console.log(key.toString());
        return [this.path, key.toString()].join("/").replace(/\/\/+/g, "/");
    }
    /**
     * Store the given value under the key.
     */
    async put(key, val) {
        try {
            await this.ossClient.put(this._getFullKey(key), buffer_1.Buffer.from(val, val.byteOffset, val.byteLength));
        }
        catch (err) {
            // auto create bucket
            if (err.status === 404 && this.createIfMissing) {
                const bucket = this.ossClient.options.bucket;
                await this.ossClient.putBucket(bucket, {
                    // https://help.aliyun.com/document_detail/32071.html
                    storageClass: "Standard",
                    acl: "public-read",
                    dataRedundancyType: "LRS",
                    timeout: 60000,
                });
                this.ossClient.useBucket(bucket);
            }
            throw Errors.dbWriteFailedError(err);
        }
    }
    /**
     * Read from oss.
     */
    async get(key) {
        try {
            const data = await this.ossClient.get(this._getFullKey(key));
            if (!data.content) {
                throw new Error("Response had no content");
            }
            // If a body was returned, ensure it's a Uint8Array
            if (data.content instanceof Uint8Array) {
                return data.content;
            }
            if (typeof data.content === "string") {
                return (0, uint8arrays_1.fromString)(data.content);
            }
            if (data.content instanceof Blob) {
                const buf = await data.content.arrayBuffer();
                return new Uint8Array(buf, 0, buf.byteLength);
            }
            return await (0, it_to_buffer_1.default)(data.content);
        }
        catch (err) {
            if (err.code === "NoSuchKey" || err.code === "NoSuchKeyError") {
                throw Errors.notFoundError(err);
            }
            throw err;
        }
    }
    /**
     * Check for the existence of the given key.
     */
    async has(key) {
        try {
            await this.ossClient.head(this._getFullKey(key));
            return true;
        }
        catch (err) {
            if (err.code === "NoSuchKey") {
                return false;
            }
            throw err;
        }
    }
    /**
     * Delete the record under the given key.
     */
    async delete(key) {
        try {
            await this.ossClient.delete(this._getFullKey(key));
        }
        catch (err) {
            throw Errors.dbDeleteFailedError(err);
        }
    }
    /**
     * Create a new batch object.
     */
    batch() {
        const puts = [];
        const deletes = [];
        return {
            put(key, value) {
                puts.push({ key: key, value: value });
            },
            delete(key) {
                deletes.push(key);
            },
            commit: async () => {
                const putOps = puts.map((p) => this.put(p.key, p.value));
                const delOps = deletes.map((key) => this.delete(key));
                await Promise.all(putOps.concat(delOps));
            },
        };
    }
    /**
     * Recursively fetches all keys from oss
     */
    async *_listKeys(params, options) {
        try {
            const data = await this.ossClient.list({
                prefix: params.prefix,
                "max-keys": 100,
                marker: params.StartAfter,
            }, {});
            if (options && options.signal && options.signal.aborted) {
                return;
            }
            if (!data || !data.objects?.length) {
                throw new Error(`Not found: ${params.prefix}`);
            }
            for (const d of data.objects) {
                if (!d.name) {
                    throw new Error(`Not found: ${params.prefix} name`);
                }
                // Remove the path from the key
                yield new interface_datastore_1.Key(d.name.slice(this.path.length), false);
            }
            // If we didn't get all records, recursively query
            if (data.isTruncated) {
                // If NextMarker is absent, use the key from the last result
                params.StartAfter = data.nextMarker;
                // recursively fetch keys
                yield* this._listKeys(params);
            }
        }
        catch (err) {
            console.log("err------.code");
            console.log(err.code);
            console.log(err);
            throw new Error(err.code);
        }
    }
    /**
     *
     */
    async *_all(q, options) {
        for await (const key of this._allKeys({ prefix: q.prefix }, options)) {
            try {
                const res = {
                    key,
                    value: await this.get(key),
                };
                yield res;
            }
            catch (err) {
                // key was deleted while we are iterating over the results
                if (err.statusCode !== 404) {
                    throw err;
                }
            }
        }
    }
    /**
     *
     */
    async *_allKeys(q, options) {
        const prefix = [this.path, q.prefix || ""].join("/").replace(/\/\/+/g, "/");
        // Get all the keys via list object, recursively as needed
        let it = this._listKeys({
            prefix: prefix,
        }, options);
        if (q.prefix != null) {
            it = (0, it_filter_1.default)(it, (k) => k.toString().startsWith(`${q.prefix || ""}`));
        }
        yield* it;
    }
    /**
     * This will check the oss bucket to ensure access and existence
     */
    async open() {
        try {
            await this.ossClient.head(this.path);
        }
        catch (err) {
            if (err.code !== "NoSuchKey") {
                throw Errors.dbOpenFailedError(err);
            }
        }
    }
    async close() { }
}
exports.OssDatastore = OssDatastore;
